Unit FileDrop;

// Thanks To the authors of DropSource.PAS:
//          Angus Johnson, ajohnson@rpi.net.au
//          Anders Melander, anders@melander.dk

Interface

Uses
  Windows, activex, Classes, shlobj, SysUtils;

Type

  pformatlist = ^tformatlist;
  tformatlist = ARRAY[0..255] Of tformatetc;

  tmydropfiles = Packed Record
    dropfiles: tdropfiles;
    filenames: ARRAY[1..1] Of Char;
  End;
  pmydropfiles = ^tmydropfiles;

  TDragType = (dtEither, dtCopy, dtMove);
  TDragResult = (drDropCopy, drDropMove, drCancel, drOutMemory, drUnknown);
  TDropEvent = Procedure(Sender: TObject; iscopynotmove: Boolean; Var continuedrop: Boolean) Of Object;

  TEnumFormatETC = Class(TInterfacedObject, IEnumFormatETC)
  Private
    fformatlist: pformatlist;
    fformatcount: Integer;
    findex: Integer;
  Public
    Constructor Create(formatlist: pformatlist; formatcount, Index: Integer);
    Function Next(celt: LongInt; Out elt; pceltfetched: plongint): hresult; StdCall;
    Function skip(celt: LongInt): hresult; StdCall;
    Function ReSet: hresult; StdCall;
    Function clone(Out enum: ienumformatetc): hresult; StdCall;
  End;

  TInterfacedComponent = Class(TObject, Iunknown)
  Private
    frefcount: Integer;
  Protected
    Function QueryInterface(Const iid: Tguid; Out obj): Integer; StdCall;
    Function _addref: Integer; StdCall;
    Function _release: Integer; StdCall;
  Public
    Property RefCount: Integer Read FRefCount;
  End;

  TDropSource = Class(TInterfacedComponent, IDropSource, IDataObject)
  Private
    FDragType: TDragType;
    FeedbackEffect: LongInt;
    FDrop: TDropEvent;
  Protected
    // IDropSource implementation
    Function querycontinuedrag(fescapepressed: bool; grfkeystate: LongInt): hresult; StdCall;
    Function givefeedback(dweffect: LongInt): hresult; StdCall;
    // IDataObject implementation
    Function getdata(Const formatetcin: tformatetc; Out medium: tstgmedium):hresult; StdCall;
    Function getdatahere(Const formatetc: tformatetc; Out medium: tstgmedium):hresult; StdCall;
    Function querygetdata(Const formatetc: tformatetc): hresult; StdCall;
    Function getcanonicalformatetc(Const formatetc: tformatetc; Out formatetcOut: tformatetc): hresult; StdCall;
    Function setdata(Const formatetc: tformatetc; Var medium: tstgmedium; frelease: bool): hresult; StdCall;
    Function enumformatetc(dwdirection: LongInt; Out enumformatetc: ienumformatetc): hresult; StdCall;
    Function dadvise(Const formatetc: tformatetc; advf: LongInt; Const advsink: iadvisesink; Out dwconnection: LongInt): hresult; StdCall;
    Function dunadvise(dwconnection: LongInt): hresult; StdCall;
    Function enumdadvise(Out enumadvise: ienumstatdata): hresult; StdCall;
    //New Functions...
    Function dogetdata(Const formatetcin: tformatetc; Out medium: tstgmedium):hresult; Virtual;
    Function dogetdatahere(Const formatetc: tformatetc; Out medium: tstgmedium):hresult; Virtual;
    Function doquerygetdata(Const formatetc: tformatetc): hresult; Virtual; Abstract;
    Function doenumformatetc(dwdirection: LongInt; Out enumformatetc: ienumformatetc): hresult; Virtual; Abstract;
  Public
    Constructor Create;
    Destructor Destroy; Override;
    Function Execute: TDragResult;
    Property DragType: TDragType Read FDragType Write FDragType;
    Property OnDrop: TDropEvent Read FDrop Write FDrop;
  End;

  TDropFileSource = Class(TDropSource, IDropSource, IDataObject)
  Private
    FFiles: TStrings;
  Protected
    Function dogetdata(Const formatetcin: tformatetc; Out medium: tstgmedium):hresult; Override;
    Function dogetdatahere(Const formatetc: tformatetc; Out medium: tstgmedium):hresult; Override;
    Function doquerygetdata(Const formatetc: tformatetc): hresult; Override;
    Function doenumformatetc(dwdirection: LongInt; Out enumformatetc: ienumformatetc): hresult; Override;
  Public
    Constructor Create;
    Destructor Destroy; Override;
    Property Files: TStrings Read FFiles;
  End;


Implementation

// -----------------------------------------------------------------------------
//			TEnumFormatETC
// -----------------------------------------------------------------------------

Constructor TEnumFormatETC.Create(formatlist: pformatlist;
  formatcount, Index: Integer);
Begin
  Inherited Create;
  fformatlist := formatlist;
  fformatcount := formatcount;
  findex := Index;
End;

Function TEnumFormatETC.Next(celt: LongInt; Out elt; pceltfetched: plongint): hresult;
Var
  i: Integer;
Begin
  i := 0;
  While (i < celt) And (findex < fformatcount) Do
  Begin
    tformatlist(elt)[i] := fformatlist[findex];
    Inc(findex);
    Inc(i);
  End;
  If pceltfetched <> Nil Then pceltfetched^ := i;
  If i = celt Then Result := s_ok
  Else Result := s_false;
End;

Function TEnumFormatETC.skip(celt: LongInt): hresult;
Begin
  If celt <= fformatcount - findex Then
  Begin
    findex := findex + celt;
    Result := s_ok;
  End
  Else
  Begin
    findex := fformatcount;
    Result := s_false;
  End;
End;

Function TEnumFormatETC.ReSet: hresult;
Begin
  findex := 0;
  Result := s_ok;
End;

Function TEnumFormatETC.clone(Out enum: ienumformatetc): hresult;
Begin
  enum := TEnumFormatETC.Create(fformatlist, fformatcount, findex);
  Result := s_ok;
End;



// -----------------------------------------------------------------------------
//			TInterfacedComponent
// -----------------------------------------------------------------------------

Function TInterfacedComponent.queryinterface(Const iid: tguid; Out obj): Integer;
Const
  e_nointerface = $80004002;
Begin
  If getinterface(iid, obj) Then Result := 0
  Else Result := e_nointerface;
End;

Function TInterfacedComponent._addref: Integer;
Begin
  Inc(frefcount);
  Result := frefcount;
End;

Function TInterfacedComponent._release: Integer;
Begin
  Dec(frefcount);
  If frefcount = 0 Then
  Begin
    destroy;
    Result := 0;
    Exit;
  End;
  Result := frefcount;
End;



// -----------------------------------------------------------------------------
//			TDropSource
// -----------------------------------------------------------------------------

Constructor TDropSource.Create;
Begin
  Inherited Create;
  _addref; //This avoids premature release!
End;

Destructor TDropSource.destroy;
Begin
  While refcount > 1 Do _release; //Allow release!
  Inherited destroy;
End;

Function TDropSource.Execute: tdragresult;
Var
  res: hresult;
  okeffect: LongInt;
  effect: LongInt;
Begin
  Result := drunknown;
  If fdragtype = dtcopy Then okeffect := dropeffect_copy
  Else If fdragtype = dtmove Then okeffect := dropeffect_move
  Else okeffect := dropeffect_copy Or dropeffect_move; //dtEither
  res := dodragdrop(Self As idataobject, Self As idropsource, okeffect, effect);
  Case res Of
    dragdrop_s_drop: Begin
                       If (okeffect And effect <> 0) Then
                       Begin
                         If (effect And dropeffect_copy <> 0) Then
                           Result := drdropcopy
                         Else Result := drdropmove;
                       End
                       Else Result := drcancel;
                     End;
    dragdrop_s_cancel: Result := drcancel;
    e_OutOfmemory:     Result := drOutmemory;
  End;
End;

Function TDropSource.querycontinuedrag(fescapepressed: bool; grfkeystate: LongInt): hresult; StdCall;
Var
  continuedrop: Boolean;
  iscopyop: Boolean;
Begin
  If (fescapepressed) Then
    Result := dragdrop_s_cancel
  Else If (grfkeystate And mk_lbutton = 0) Then
  Begin
    continuedrop := True;
    iscopyop := True;
    If (feedbackeffect And dropeffect_copy <> 0) Then //IsCopyOp := true
    Else If (feedbackeffect And dropeffect_move <> 0) Then iscopyop := False
    Else continuedrop := False;

    //If a valid drop Then Do OnDrop event If assigned...
    If continuedrop And ((iscopyop And (dragtype <> dtmove)) Or
      (NOT iscopyop And (dragtype <> dtcopy))) And
      assigned(ondrop) Then ondrop(Self, iscopyop, continuedrop);

    If continuedrop Then Result := dragdrop_s_drop
    Else Result := dragdrop_s_cancel;
  End
  Else Result := noerror;
End;

Function TDropSource.givefeedback(dweffect: LongInt): hresult; StdCall;
Begin
  feedbackeffect := dweffect;
  Result:=dragdrop_s_usedefaultcursors;
End;

Function TDropSource.getcanonicalformatetc(Const formatetc: tformatetc; Out formatetcOut: tformatetc): hresult;
Begin
  formatetcOut.ptd := Nil;
  Result := e_notimpl;
End;

Function TDropSource.setdata(Const formatetc: tformatetc; Var medium: tstgmedium; frelease: bool): hresult;
Begin
  Result := e_notimpl;
End;

Function TDropSource.dadvise(Const formatetc: tformatetc; advf: LongInt; Const advsink: iadvisesink; Out dwconnection: LongInt): hresult;
Begin
  Result := ole_e_advisenotsupported;
End;

Function TDropSource.dunadvise(dwconnection: LongInt): hresult;
Begin
  Result := ole_e_advisenotsupported;
End;

Function TDropSource.enumdadvise(Out enumadvise: ienumstatdata): hresult;
Begin
  Result := ole_e_advisenotsupported;
End;

Function TDropSource.getdata(Const formatetcin: tformatetc; Out medium: tstgmedium):hresult; StdCall;
Begin
  Result := dogetdata(formatetcin, medium);
End;

Function TDropSource.getdatahere(Const formatetc: tformatetc;
  Out medium: tstgmedium):hresult; StdCall;
Begin
  Result := dogetdatahere(formatetc, medium);
End;

Function TDropSource.querygetdata(Const formatetc: tformatetc): hresult; StdCall;
Begin
  Result := doquerygetdata(formatetc);
End;

Function TDropSource.enumformatetc(dwdirection: LongInt; Out enumformatetc:ienumformatetc): hresult; StdCall;
Begin
  Result := doenumformatetc(dwdirection, enumformatetc);
End;

Function TDropSource.dogetdata(Const formatetcin: tformatetc; Out medium: tstgmedium):hresult;
Begin
  Result := dv_e_formatetc;
End;

Function TDropSource.dogetdatahere(Const formatetc: tformatetc; Out medium: tstgmedium):hresult;
Begin
  Result := dv_e_formatetc;
End;



// -----------------------------------------------------------------------------
//			TDropFileSource
// -----------------------------------------------------------------------------

Constructor TDropFileSource.Create;
Begin
  Inherited Create;
  ffiles := tstringlist.Create;
End;

Destructor TDropFileSource.destroy;
Begin
  ffiles.Free;
  Inherited destroy;
End;

Function TDropFileSource.dogetdatahere(Const formatetc: tformatetc; Out medium: tstgmedium):hresult;
Var
  i: Integer;
  dropfiles: pmydropfiles;
  ptext: PChar;
Begin
  If (formatetc.cfformat = cf_hdrop) And
    ((formatetc.dwaspect = dvaspect_content) Or (formatetc.dwaspect = dvaspect_icon)) And
    (formatetc.tymed = tymed_hglobal) And (medium.tymed = tymed_hglobal) Then
  Begin
    If (medium.hglobal = 0) Then
    Begin
      Result:=e_OutOfmemory;
      Exit;
    End;
    dropfiles := globallock(medium.hglobal);
    Try
      dropfiles^.dropfiles.pfiles := SizeOf(tdropfiles);
      dropfiles^.dropfiles.fwide := False;
      ptext := @(dropfiles^.filenames);
      For i := 0 TO ffiles.Count-1 Do
      Begin
        StrPCopy(ptext, ffiles[i]);
        Inc(ptext, Length(ffiles[i])+1);
      End;
      ptext^ := #0;
    Finally
      globalunlock(medium.hglobal);
    End;
    medium.unkforrelease := Nil;
    Result := s_ok;
  End
  Else Result := Inherited dogetdatahere(formatetc, medium);
End;

Function TDropFileSource.dogetdata(Const formatetcin: tformatetc; Out medium: tstgmedium):hresult;
Var
  i: Integer;
  strlength: Integer;
Begin
  medium.tymed := 0;
  medium.unkforrelease := Nil;
  medium.hglobal := 0;

  If (formatetcin.cfformat = cf_hdrop) And
    ((formatetcin.dwaspect = dvaspect_content) Or (formatetcin.dwaspect = dvaspect_icon)) And
    (formatetcin.tymed = tymed_hglobal) Then
  Begin
    strlength := 0;
    For i := 0 TO ffiles.Count-1 Do
      Inc(strlength, Length(ffiles[i])+1);
    medium.hglobal := globalalloc(gmem_share Or gmem_zeroinit, SizeOf(tdropfiles)+strlength+1);
    If (medium.hglobal = 0) Then
    Begin
      Result:=e_OutOfmemory;
      Exit;
    End;
    medium.tymed := tymed_hglobal;
    Result := dogetdatahere(formatetcin, medium);
  End
  Else Result := dv_e_formatetc;
End;

Function TDropFileSource.doquerygetdata(Const formatetc: tformatetc): hresult;
Begin
  // This method is called by the drop target to check whether the source
  // provides data in a format that the target accepts.
  If (formatetc.cfformat = cf_hdrop) And
    ((formatetc.dwaspect = dvaspect_content) Or (formatetc.dwaspect = dvaspect_icon)) And
    (formatetc.tymed = tymed_hglobal) Then Result:=s_ok
  Else Result:=s_false;
End;

Const
  filedataformatcount = 2;

Var
  filedataformats: ARRAY[0..filedataformatcount - 1] Of tformatetc;

Function TDropFileSource.doenumformatetc(dwdirection: LongInt; Out enumformatetc:ienumformatetc): hresult;
Begin
  If (dwdirection = datadir_get) Then
  Begin
    filedataformats[0].cfformat := cf_hdrop;
    filedataformats[0].dwaspect := dvaspect_content;
    filedataformats[0].lindex := -1;
    filedataformats[0].tymed := tymed_hglobal;
    filedataformats[1].cfformat := cf_hdrop;
    filedataformats[1].dwaspect := dvaspect_icon;
    filedataformats[1].lindex := -1;
    filedataformats[1].tymed := tymed_hglobal;
    enumformatetc := TEnumFormatETC.Create(@filedataformats, filedataformatcount, 0);
    Result := s_ok;
  End
  Else If (dwdirection = datadir_set) Then
    Result := e_notimpl
  Else Result := e_invalidarg;
End;



Initialization
  OleInitialize(Nil);

Finalization
  OleUninitialize;

End.
